/*
 * database.h - a C++ class for disk databases using gdbm
 *
 * This is a class to store assignments
 *   (key) -> (value)
 * in a database that is kept on disk (and hence survives program restarts
 * if the programmer wants this). Key and value can be of any type. To open a
 * database, write
 *   database db("filename","directory",timestamp);
 * where the last two arguments may be omitted. This doesn't open the file yet,
 * but it creates the directory if it doesn't exist. If CHECK_TIME is defined,
 * the database is deleted if it is older than the timestamp passed. (This
 * allows for automatic recreation of the database when the program changes;
 * the timestamp passed should be the compilation time of the program.)
 *
 * The file is opened automatically upon the first read/write request, or
 * with the explicit
 *   db.open ();
 * call. The database is synced after every write operation, and closed
 * automatically on program exit (with the database destructor), or with the
 *   db.close ();
 * call. The database can be cleared with
 *   db.rm ();
 * To write an entry into the database, you have to set the key and the value.
 * The key can be set with
 *   db.reset_key ();
 *   (... statements to insert data into the key ...)
 * The most basic statements to insert data are
 *   db.set_key_bytes (char *s, int n);
 * and
 *   db.set_key_bits (int x, int n);
 * which inserts n characters from s, and the least significant n bits from x,
 * respecively. The two functions can be mixed, however the bit function must
 * always write full bytes before the byte function (or anything else) can be
 * called again. The better way is probably to say
 *   set_key (db, x);
 * to insert x, where x is of "any" type. In this library, this is predefined
 * for char, strings, and int. Ints are stored in a very efficient way, see
 * database.cc for details on this. Libraries providing other classes
 * should come with their own set_key (database, type) function. The value can
 * be set in the same way. Multiple set_key and set_value statements are
 * allowed. At the end, you have to call db.write (). For example, a complete
 * write operation might look like this:
 *   db.reset_key ();
 *   db.reset_value ();
 *   set_key (db,"Hi!");
 *   set_value (db,42);
 *   set_value (db,'x');
 *   db.write ();
 * This stores the assignment ("Hi") -> (42,'x'). If an entry "Hi" already
 * existed in the database, it is replaced.
 *
 * Reading an entry works in an analogous way. To read the above entry, write
 *   db.reset_key ();
 *   set_key (db,"Hi!");
 *   assert ( db.read () ); // otherwise the key is not in the database
 *   int i; char c;
 *   get_value (db,i);
 *   get_value (db,c);
 * which stores 42 in i and 'x' in c. It is up to the programmer to remember
 * which types have been used for key and value - using different types for
 * the same entry will give unpredictable results.
 *
 *
 * Note:
 * - The final program has to be linked with -lgdbm.
 * - The key and value data are built up in an internal memory area. There
 *   are 256 static bytes predefined for this. Longer keys and values are
 *   allowed, in which case more memory is allocated. Usually (for keys and
 *   values not exceeding length 256), no memory requests are made. There is
 *   never a memory request for reading from a database, as the corresponding
 *   gdbm_fetch function allocates memory itself. This memory is freed upon
 *   the next read or call to reset_value.
 * - The number of bytes used to store int's is variable and depends on the
 *   size of the actual integer. In general, about n/7 bytes of storage are
 *   needed for an integer whose dual representation has n digits. In
 *   particular, "small" numbers will need only 1 byte, which can lead to
 *   great memory savings over storing full sizeof(int) bytes for every
 *   integer. See database.cc for more details.
 * - The way int's are stored is independent of the endianness of the machine.
 * - The database is currently never reorganized. This should be no big
 *   problem if entries are never overwritten.
 *
 */

#ifndef __DATABASE_H__
#define __DATABASE_H__

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>
#include <string.h>
#include <gdbm.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

class database {
	public:
		database (char *, char * = NULL, time_t = 0);
		~database () {
			if (op) close ();
			delete [] name;
		};

		bool open ();
		bool close ();
		bool rm (); // remove database from disk

		void reset_key () {
			key.dsize = 0;
			if (key.dptr != key_space) delete [] key.dptr;
			key.dptr = key_space;
			key_alloc = 256;
			key_cur = 0;
			key_pos = 0;
		};
		void reset_value () {
			value.dsize = 0;
			if (value.dptr != value_space) delete [] value.dptr;
			value.dptr = value_space;
			value_alloc = 256;
			value_cur = 0;
			value_pos = 0;
		};
		// Byte-wise writing/reading; second arg is number of bytes
		void set_key_bytes (char *, int);
		void set_value_bytes (char *, int);
		bool get_value_bytes (char *&, int);
		// Bit-wise reading/storing; second arg is number of bits
		void set_key_bits (int, int);
		void set_value_bits (int, int);
		bool get_value_bits (int&, int);
		// Write and read access
		bool write ();
		bool read ();

		// Additional set/get_key/value functions for more data types
		friend void set_key (database&, char *);
		friend void set_value (database&, char *);
		friend bool get_value (database&, char *&);
		friend void set_key (database&, char);
		friend void set_value (database&, char);
		friend bool get_value (database&, char&);
		friend void set_key (database&, int);
		friend void set_value (database&, int);
		friend bool get_value (database&, int&);
	private:
		char *name;
		char *dir;
		GDBM_FILE dbm;
		int op; // is the database open?

		datum key, value; // the data structure for the gdbm functions
		char key_space[256];
		char value_space[256];
		int key_alloc, value_alloc; // allocated memory
		int value_read; // read position for value
		char key_cur, value_cur; // current byte for bit-wise storing
		char key_pos, value_pos; // current bit for bit-wise storing.
					 // Possible values are 0 to 7.
};

#endif

