/*
 * linsys.h - a class for solving linear systems of equations in Q
 *
 * We solve linear systems of equations by application of the Gauss algorithm.
 * New equations and variables can be added at any time; the system constantly
 * checks whether the system is consistent and uniquely solvable.
 *
 * To initialize a system of equations, write
 *   linsys Q(n);
 * where n is the maximum number of variables that may occur.
 *
 * Variables are labeled by a name (any integer). To introduce a new variable
 * without specifying an equation right now, say
 *   Q.add_var (name);
 * This function returns -1 if there are too many variables, or a non-negative
 * integer, which is then the internal number of the variable. Note that this
 * internal number may change while solving the system, so it's probably not
 * useful. Usually it is not necessary to introduce variables this way.
 *
 * To build up an equation, you can add terms by writing
 *   Q.add_var_term (name, coeff);
 * where name is the name of the variable as above, and coeff the coefficient
 * (a rational number). This adds a term coeff*x[name] to the left hand side of
 * the equation. Multiple calls for the same variable are allowed; the
 * coefficients are added up then. If the variable doesn't exist yet, it is
 * introduced. The function returns -1 on error (too many variables) or a
 * non-negative integer otherwise (the internal number of the variable). To
 * add a constant term to the right hand side of the equation, use
 *   Q.add_const_term (coeff);
 * in the same way. Once the equation is built up completely, you can store it
 * with
 *   Q.send_row();
 * This function returns a linsys_t which is
 * - CONTRADICTION if the equations given so far are not solvable,
 * - UNDETERMINED if the equations don't give a contradiction, but they are
 *     not sufficient yet to determine the variables,
 * - SOLVABLE if the equations determine the variables uniquely.
 * The caller MUST check if the result is CONTRADICTION and stop using the
 * linear system any further in this case. The caller may add further equations
 * (and variables) if the system is solvable already.
 *
 * If the system is solvable, you can get the result by the following
 * functions:
 * - Q.value_by_name (name) is the result for x[name],
 * - Q.value_by_number (number) is the result for the variable with the
 *   specified internal number. The number of variables is Q.numvars(), and the
 *   name of the variable with a given internal number is Q.varname(number).
 *
 */

#ifndef __LINSYS_H__
#define __LINSYS_H__

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>
#include <gmp.h>

#include "rational.h"



enum linsys_t {CONTRADICTION, UNDETERMINED, SOLVABLE};

class linsys {
	public:
		linsys (int size) {
			maxsize = size;
			cols = rows = 0;
			assert (varnames = new int[size]);
			assert (entries = new rational[(size+1)*(size+1)]);
			entry (0,size) = 0;
		};
		~linsys () {
			delete [] varnames;
			delete [] entries;
		};

		int add_var (int);
		int add_var_term (int v, rational x) {
			int i = add_var (v);
			if (i>=0) entry (rows,i) += x;
			return i;
		};
		void add_const_term (rational x) {entry (rows,maxsize) += x;};
		enum linsys_t send_row ();
		int numvars () {return cols;};
		int varname (int i) {return varnames[i];};
		rational value_by_name (int);
		rational value_by_number (int);
		void write_current_eqn ();
		rational det(); // return the determinant of the system
				// in case it is solvable

		friend ostream& operator<< (ostream&, const linsys&);

	private:
		int maxsize;
		int cols,rows;

		/*
		 * The system is always stored in the form
		 *                rows    cols   maxsize
		 *                 |       |      |
		 *                 V       V      V
		 *         X 0 0 0 * * * *        *
		 *         0 X 0 0 * * * *        *
		 *         0 0 X 0 * * * *        *
		 *         0 0 0 X * * * *        *
		 * rows->
		 * where X is a non-zero and * any entry. Dependent equations
		 * are not stored. The system is solvable if rows==cols.
		 *
		 */
		int *varnames; // names of the variables from 0 to cols-1
		rational *entries; // the array indicated above

		rational& entry(int r, int c) const
			{return entries [(maxsize+1)*c+r];};
};

#endif

