/*
 * trr.h - implementing the topological recursion relations for any genus
 *
 * Currently we cannot deal with primitive classes; the restrictions are
 * essentially inherited from split2.
 *
 * To speed up the computations, we introduce new invariants for curves of
 * genus 0. Internally, these invariants are stored with a negative genus
 * g=-k for k>0. Such an invariant will always have at least 2 marked points.
 * The first two marked points x_1 and x_2 are special (and will never be
 * moved when sorting the invariant). The invariant is simply a sum of all
 * chain graphs
 *
 *          X       X       X
 *         / \     / \     / \
 *        /   \   /    ...    \
 *  x_1  /     \ /             \  x_2
 *      X       X               X
 *
 * where:
 * - every component has genus 0,
 * - the _left_ node of every component (including x_1) can get any number of
 *   psi classes, but such that:
 * - the number of nodes plus the number of added psi classes equals k,
 * - the diagonal splitting must be observed at every node when forgetting
 *   the psi classes,
 * - the degree and the remaining marked points are distributed in all
 *   possible ways onto the components,
 * - every graph is weighted with the coefficient (-1)^(number of nodes).
 * The _right_ node of every component (including x_2) does not have psi
 * classes.
 *
 * Note that in particular we get ordinary genus-0 invariants for k=g=0.
 * The moduli space for all such curves has pure virtual codimension k in the
 * moduli space of genus-0 curves.
 *
 * Although these invariants contain splittings into an arbitrary number of
 * components, we can compute them recursively using only split2 (i.e.
 * splittings into two components). Namely,
 * - for k=0 the invariants are ordinary genus-0 GW invariants that we have
 *   already computed.
 * - for k>0, we have
 *     I_k (alpha_1,alpha_2,...) =
 *         I_{k-1} (psi*alpha_1,alpha_2,...)
 *       - I_0 (alpha_1,X,...) * I_{k-1} (X,alpha_2,...)
 *   where X denotes a diagonal splitting, and the .../... in the last line
 *   denotes every possible (split2) splitting of degree and classes.
 *
 *   Note that the first entry alpha_1 may have psi classes, but the second
 *   alpha_2 may not. In the last line of the above recursion we're not
 *   using psi classes in the first entry of the second factor, but that's OK.
 *   In the above formulas it may happen that we get a 2-point rational
 *   degree-0 invariant; of course this is then defined to be zero.
 *
 *   Note moreover that the first invariant in the last line above is an
 *   ordinary genus-0 invariant where all points can be sorted. So we can put
 *   the diagonal class in the first entry, which is then consistent with the
 *   diagonal summing of the split2 functions.
 *
 * Using the above invariants, the topological recursion relation is now easily
 * stated:
 *   I_g (alpha_1*psi^{3g-1},alpha_2,...) =
 *     sum I_k (alpha_1,X,...) * I_g (X*psi^l,...)
 * where the sum is taken over all splitings of the degree and the classes,
 * over the diagonal, and over k+l=3g-2 (k>=0 and l>=0). alpha_1 may itself
 * contain psi classes. Note that, unlike all other cases considered so far,
 * there are psi classes on the diagonal.
 *
 */

#ifndef __TRR_H__
#define __TRR_H__

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>

#include "gwinv.h"
#include "database.h"
#include "linsys.h"
#include "compute.h"



// Reduce an invariant modulo the topological recursion relations. Some
// invariants may be returned as variables. See trr.cc and virasoro.cc for
// more details.

rational trr_variable (linsys&, inv&, rational&, int);

// Compute an invariant using the topological recursion relations, if
// possible.

rational trr_genusg (inv&, int);

#endif

