/*
 * virasoro.h - computing GW invariants using the Virasoro conditions
 *
 * Currently we cannot deal with primitive classes; the restrictions are
 * essentially inherited from split2.
 *
 * Currently (and probably on principle) this works only for Fano varieties.
 * Actually, in the current version only P^n is supported. This has three
 * reasons:
 *  1) in the (C) and (D) terms below, we need an xdim (or 1/xdim) factor due
 *     to raising/lowering indices. (This is certainly easily fixed.)
 *  2) in the (C) terms below, primitive classes will contribute! This has
 *     not been implemented yet.
 *  3) The Virasoro conjecture has not been proven yet for varieties other
 *     than P^n.
 *
 * Fix an integer k>=-1 and a genus g>=1. The Virasoro condition L_k reads
 *   0 = (A)+(B)+(C)+(D)+(E)
 * where
 *   (A) = - sum_{i=0}^{k+1} symfunc((3-r)/2,k,i) (R^i)_0^b
 *             << tau_{k+1-i,b} >>_g
 *   (B) = sum_{i=0}^{k+1} sum_{m=0}^infty symfunc (mu_a+m+1/2,k,i)
 *             (R^i)_a^b t_m^a << tau_{k+m-i,b} >>_g
 *   (C) = 1/2 sum_{i=0}^{k+1} sum_{m=i-k}^{-1} symfunc (-mu_a+m+1/2,k,i)
 *             (-1)^m (R^i)^{ab} << tau_{-m-1,a} tau_{m+k-i,b} >>_{g-1}
 *   (D) = 1/2 sum_{i=0}^{k+1} sum_{m=i-k}^{-1} symfunc (-mu_a+m+1/2,k,i)
 *             (-1)^m (R^i)^{ab} sum_{h=0}^g
 *             << tau_{-m-1,a} >>_h << tau_{m+k-i,b} >>_{g-h}
 *   (E) = 1/48 delta_{k,0} delta_{g,1}
 *             int_X ( (3-r)c_r(X) - 2c_1(X)c_{r-1}(X) )
 * Here, r=dim X, mu_a = p_a - r/2 if the class named a has Hodge type (p,q),
 * and R is the matrix of multiplication with c_1(X). If we restrict to k>0
 * we can obviously forget (E). Note that the sums over i are also bounded
 * above by r because of the (R^i) factor.
 *
 * Here is an example: consider k=2 and g=1 in P^3. Moreover, consider the
 * t-power corresponding to the one class H^3, and degree 1. Then the terms
 * are (where we write invariants as I_{g,d} (...):
 *   (A) = -symfunc(0,2,0)*I_{1,1}(psi^3,H^3)              0
 *         -4*symfunc(0,2,1)*I_{1,1}(H*psi^2,H^3)          -8 * 1/6        -4/3
 *         -16*symfunc(0,2,2)*I_{1,1}(H^2*psi,H^3)         -48 * -1/6      8
 *         -64*symfunc(0,2,3)*I_{1,1}(H^3,H^3)             -64 * -1/12     16/3
 *   (B) = symfunc(2,2,0)*I_{1,1}(H^3*psi^2)               24 * -1/12      -2
 *   (C) = 1/2 symfunc(-3,2,0)*I_{0,1}(H^3*psi,1,H^3)      -3 * 1          -3
 *         -1/2 symfunc(-2,2,0)*I_{0,1}(H^3,psi,H^3)       0
 *         1/2 symfunc(-2,2,0)*I_{0,1}(H^2*psi,H,H^3)      0
 *         -1/2 symfunc(-1,2,0)*I_{0,1}(H^2,H*psi,H^3)     0
 *         1/2 symfunc(-1,2,0)*I_{0,1}(H*psi,H^2,H^3)      0
 *         -1/2 symfunc(0,2,0)*I_{0,1}(H,H^2*psi,H^3)      0
 *         1/2 symfunc(0,2,0)*I_{0,1}(psi,H^3,H^3)         0
 *         -1/2 symfunc(1,2,0)*I_{0,1}(1,H^3*psi,H^3)      -3 * 1          -3
 *
 *         -4/2 symfunc(-2,2,1)*I_{0,1}(H^3,H,H^3)         -4 * 1          -4
 *         -4/2 symfunc(-1,2,1)*I_{0,1}(H^2,H^2,H^3)       2 * 1           2
 *         -4/2 symfunc(0,2,1)*I_{0,1}(H,H^3,H^3)          -4 * 1          -4
 *   (D) = -symfunc(0,2,0)*I_{1,0}(H) I_{0,1}(H^2*psi,H^3) 0
 *         symfunc(0,2,0)*I_{1,0}(psi) I_{0,1}(H^3,H^3)    0
 *         -4*symfunc(0,2,1)*I_{1,0}(H) I_{0,1}(H^3,H^3)   -8 * -1/4       2
 *                                                                        -----
 *                                                                         0
 *
 * We'll compute the invariants first of all by induction on the genus, so
 * (C) can be considered to be known, as well as all terms of (D) where h is
 * not equal to 0 or g. Next, we'll do induction on the number of marked
 * points, so (B) can be considered to be known too, as well as the (D) terms
 * where not all free marked points are on the genus-g component. The
 * remaining terms differ only by the degree and the class on the first marked
 * point. We consider all these invariants (there are only finitely many in
 * the Fano case) as unknowns, and consider the above conditions for one k
 * after the other (there is one such useful equation for every degree of the
 * curves) until we have enough linearly independent equations. Here, the
 * non-recursively-known genus-g invariants of (A) and (D) have to be reduced
 * immediately using the general topological recursion relations.
 *
 */

#ifndef __VIRASORO_H__
#define __VIRASORO_H__

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>

#include "gwinv.h"
#include "database.h"
#include "compute.h"



// Compute an invariant using the Virasoro conditions, if possible.

rational virasoro (inv&, int);

#endif

